"""
Session adapters

"""
import logging
import os
from os import path
import tempfile
import time
import uuid

from tinman import config
from tinman import exceptions
from tinman import serializers

LOGGER = logging.getLogger(__name__)

DEFAULT_DURATION = 300


class SessionAdapter(object):
    """Base session adapter that facilitates the storage and retrieval of
    session data.

    """
    def __init__(self, application, session_id=None, configuration=None,
                 duration=DEFAULT_DURATION, serializer=None):
        """Create a session adapter for the base URL specified, creating a new
        session if no session id is passed in.

        :param str session_id: The current session id if once has been started
        :param dict configuration: Session storage configuration
        :param int duration: The session duration for storage purposes
        :type serializer: tinman.session.serializer.SessionSerializer

        """
        self.__dict__['attributes'] = dict()
        self.__dict__['attributes']['_id'] = session_id or self._create_id()
        self.__dict__['data'] = dict()
        self._application = application
        self._config = configuration or dict()
        self._duration = duration
        self._serializer = serializer or serializers.Pickle()

    def __contains__(self, item):
        """Check to see if the item is in the session data dictionary.

        :param str item: The item to look for

        """
        return item in self.__dict__['data']

    def __delattr__(self, key):
        """Delete a session object's attribute, keeping track of what is
        internal to the object itself and what is a data attribute.

        :param str key: The attribute to remove
        :raises: AttributeError

        """
        if key == '_data':
            self.__dict__['data'] = dict()
        elif key[0] == '_':
            if key in self.__dict__['attributes']:
                del self.__dict__['attributes'][key]
        elif key in self.__dict__['data']:
            del self.__dict__['data'][key]
        else:
            raise AttributeError(key)

    def __getattr__(self, key):
        """Get a session object's attribute, keeping track of what is internal
        to the object and what is a session data attribute.

        :param str key: The key to get the value for
        :rtype: any

        """
        if key == '_data':
            return self.__dict__['data']
        elif key[0] == '_':
            return self.__dict__['attributes'].get(key)
        return self.__dict__['data'].get(key)

    def __hash__(self):
        """Return the hash of the object, using the hash of the session id.

        :rtype: int

        """
        return self.__dict__['attributes']['_id'].hash()

    def __iter__(self):
        """Iterate through the keys in the data dictionary.

        :rtype: list

        """
        return iter(self.__dict__['data'])

    def __len__(self):
        """Return the length of the data dictionary.

        :rtype: int

        """
        return len(self.__dict__['data'])

    def __repr__(self):
        """Return the representation of the class as a string.

        :rtype: str

        """
        return '<%s(%r)>' % (self.__class__.__name__, self.id)

    def __setattr__(self, key, value):
        """Set a session object's attribute, keeping track of what is internal
        to the object and what is a session data attribute.

        :param str key: The attribute name
        :param any value: The value to set

        """
        if key == '_data':
            raise AttributeError('Can not set the _data value')
        if key[0] == '_':
            self.__dict__['attributes'][key] = value
        else:
            self.__dict__['data'][key] = value

    def _create_id(self):
        """Create a session id

        :rtype: str

        """
        return str(uuid.uuid4())

    def _deserialize(self, data):
        """Return the deserialized session data.

        :rtype: str

        """
        return self._serializer.deserialize(data)

    def _load_session_data(self):
        """Extend to the load the session from storage

        :raises: NotImplementedError

        """
        raise NotImplementedError

    def _serialize(self):
        """Return the session data as serialized string.

        :rtype: str

        """
        return self._serializer.serialize(self.__dict__['data'])

    def as_dict(self):
        """Return the SessionAdapter object as a dictionary value, returning
        only the session data attributes in the dictionary.

        :rtype: dict

        """
        return self.__dict__['data']

    def clear(self):
        """Clear all of the attributes in the session."""
        self._data = dict()

    def delete(self):
        """Extend to the delete the session from storage

        :raises: NotImplementedError

        """
        raise NotImplementedError

    def get(self, item, default):
        """Get an item from the session data returning a default value if it
        is not present.

        :param str item: The item in the session data to get
        :param any default: The default value
        :rtype: any

        """
        return self.__dict__['data'].get(item, default)

    @property
    def id(self):
        """A read-only view of the session id.

        """
        return self.__dict__['attributes']['_id']

    def items(self):
        """Return the session data items as a list of key/value tuples.

        :rtype: list

        """
        return self.__dict__['data'].items()

    def keys(self):
        """Return the session data keys

        :rtype: list

        """
        return self.__dict__['data'].keys()

    def load(self):
        """Load the session data"""
        self.__dict__['data'] = self._load_session_data()

    def save(self):
        """Save the session for later retrieval

        :raises: NotImplementedError

        """
        raise NotImplementedError

    def values(self):
        """Return the session data values

        :rtype: list

        """
        return self.__dict__['data'].values()


class FileSessionAdapter(SessionAdapter):
    """Stores session data on the filesystem. Configuration values that can
    be set:

    directory: The directory to store the files in
    cleanup: Boolean value to clean stale session files

    If directory is not set, a directory named "tinman" will be created in the
    default system temp directory for the active user and files will be placed
    in there.

    Cleanup is turned on by default

    """
    SUBDIR = 'tinman'

    def __init__(self, application, session_id=None, configuration=None,
                 duration=DEFAULT_DURATION, serializer=None):
        """Create a session adapter for the base URL specified, creating a new
        session if no session id is passed in.

        :param tinman.application.Application application: The Tinman app
        :param str session_id: The current session id if once has been started
        :param dict configuration: Session storage configuration
        :param int duration: The session duration for storage retention
        :param serializer: The object to use for session serialization
        :type serializer: tinman.session.serializer.SessionSerializer
        """
        super(FileSessionAdapter, self).__init__(application, session_id,
                                                 configuration, duration,
                                                 serializer)
        LOGGER.info(self._serializer)
        self._storage_dir = self._setup_storage_dir()
        if self._config.get('cleanup', True):
            self._cleanup()

    def _cleanup(self):
        """Remove any stale files from the session storage directory"""
        for filename in os.listdir(self._storage_dir):
            file_path = path.join(self._storage_dir, filename)
            file_stat = os.stat(file_path)
            evaluate = max(file_stat.st_ctime, file_stat.st_mtime)
            if evaluate + self._duration < time.time():
                LOGGER.debug('Removing stale file: %s', file_path)
                os.unlink(file_path)

    @property
    def _default_path(self):
        """Return the default path for session data

        :rtype: str

        """
        return path.join(tempfile.gettempdir(), self.SUBDIR)

    @property
    def _filename(self):
        """Returns the filename for the session file.

        :rtype: str

        """
        return path.join(self._storage_dir, self.id)

    def _load_session_data(self):
        """Extend to the load the session from storage

        :rtype: dict
        :raises: IOError

        """
        try:
            with open(self._filename, 'rb') as session_file:
                return self._deserialize(session_file.read()) or dict()
        except IOError as error:
            LOGGER.debug('Session file error: %s', error)
        return dict()

    def _make_path(self, dir_path):
        """Create the full path specified.

        :param str dir_path: The path to make

        """
        os.makedirs(dir_path, 0x755)

    def _setup_storage_dir(self):
        """Setup the storage directory path value and ensure the path exists.

        :rtype: str
        :raises: tinman.exceptions.ConfigurationException

        """
        dir_path = self._config.get('directory')
        if dir_path is None:
            dir_path = self._default_path
            if not os.path.exists(dir_path):
                self._make_path(dir_path)
        else:
            dir_path = path.abspath(dir_path)
            if not os.path.exists(dir_path) or not os.path.isdir(dir_path):
                raise exceptions.ConfigurationException('FileSessionAdapter '
                                                        'directory')
        return dir_path.rstrip('/')

    def delete(self):
        """Remove the session data from storage, clearing any session values"""
        self.clear()
        if os.path.isfile(self._filename):
            os.unlink(self._filename)
        else:
            LOGGER.debug('Session file did not exist: %s', self._filename)

    def save(self):
        """Store the session for later retrieval

        :raises: IOError

        """
        try:
            with open(self._filename, 'wb') as session_file:
                session_file.write(self._serialize())
        except IOError as error:
            LOGGER.debug('Session file error: %s', error)
            raise error


def get_session_serializer(configuration):
    """Return a data serializer for use with the session adapter for the given
    configuration.
    :param dict configuration: Session configuration
    :rtype: tinman.serializer.Serializer

    """
    serializer = configuration.get('serializer')
    if serializer == 'json':
        return serializers.JSON()
    elif serializer == 'msgpack':
        return serializers.MsgPack()
    return serializers.Pickle()


def get_session_adapter(application, session_id, configuration, duration):
    """Return a new instance of a session adapter for the given application,
    session, configuration and duration.

    :param tornado.web.Application application: The Tornado application
    :param str session_id: The session id
    :param dict configuration: Session adapter configuration
    :param int duration: Session duration
    :rtype: tinman.session.SessionAdapter

    """
    serializer = get_session_serializer(configuration)
    if configuration.get('name') == config.FILE:
        return FileSessionAdapter(application,
                                  session_id,
                                  configuration,
                                  duration,
                                  serializer)

    elif configuration.get('name') == config.REDIS:
        from tinman.session import redis_adapter
        return redis_adapter.RedisSessionAdapter(application,
                                                 session_id,
                                                 configuration,
                                                 duration,
                                                 serializer)

    raise exceptions.ConfigurationException('Session Adapter')



"""The RabbitMQRequestHandler wraps RabbitMQ use into a request handler, with
methods to speed the development of publishing RabbitMQ messages.

Example configuration:

    Application:
      rabbitmq:
        host: rabbitmq1
        virtual_host: my_web_app
        username: tinman
        password: tornado

"""
import logging
import pika
from pika.adapters import tornado_connection
from tornado import web

LOGGER = logging.getLogger(__name__)

from tinman import exceptions

message_stack = list()
pending_rabbitmq_connection = None
rabbitmq_connection = None


class RabbitMQRequestHandler(web.RequestHandler):
    """The request handler will connect to RabbitMQ on the first request,
    buffering any messages that need to be published until the Channel to
    RabbitMQ is opened, sending the stack of previously buffered messages at
    that time. If RabbitMQ closes it's connection to the app at any point, a
    connection attempt will be made on the next request.

    Expects configuration in the YAML file under a "rabbitmq" node. All of the
    configuration values are optional but username and password:

        host: Hostname, defaults to localhost if omitted
        port: RabbitMQ port, defaults to 5672 if omitted
        virtual_host: The virtual host, defaults to / if omitted
        username: The username to connect with
        password: The password to connect with
        channel_max: Maximum number of channels to allow, defaults to 0
        frame_max: The maximum byte size for an AMQP frame, defaults to 131072
        heartbeat_interval: Heartbeat interval, defaults to 0 (Off)
        ssl: Enable SSL, defaults to False
        ssl_options: Arguments passed to ssl.wrap_socket as described at
                     http://docs.python.org/dev/library/ssl.html
        connection_attempts: Maximum number of retry attempts, defaults to 1
        retry_delay: Time to wait in seconds between attempts, defaults to 2
        socket_timeout: Use for high latency networks, defaults to 0.25
        locale: Set the connection locale value, defaults to en_US

    """
    CHANNEL = 'rabbitmq_channel'
    CONNECTION = 'rabbitmq_connection'

    def _add_to_publish_stack(self, exchange, routing_key, message, properties):
        """Temporarily add the message to the stack to publish to RabbitMQ

        :param str exchange: The exchange to publish to
        :param str routing_key: The routing key to publish with
        :param str message: The message body
        :param pika.BasicProperties: The message properties

        """
        global message_stack
        message_stack.append((exchange, routing_key, message, properties))

    def _connect_to_rabbitmq(self):
        """Connect to RabbitMQ and assign a local attribute"""
        global pending_rabbitmq_connection, rabbitmq_connection
        if not rabbitmq_connection:
            LOGGER.info('Creating a new RabbitMQ connection')
            pending_rabbitmq_connection = self._new_rabbitmq_connection()

    def _new_message_properties(self, content_type=None, content_encoding=None,
                                headers=None, delivery_mode=None, priority=None,
                                correlation_id=None, reply_to=None,
                                expiration=None, message_id=None,
                                timestamp=None, message_type=None, user_id=None,
                                app_id=None):
        """Create a BasicProperties object, with the properties specified

        :param str content_type: MIME content type
        :param str content_encoding: MIME content encoding
        :param dict headers: Message header field table
        :param int delivery_mode: Non-persistent (1) or persistent (2)
        :param int priority: Message priority, 0 to 9
        :param str correlation_id: Application correlation identifier
        :param str reply_to: Address to reply to
        :param str expiration: Message expiration specification
        :param str message_id: Application message identifier
        :param int timestamp: Message timestamp
        :param str message_type: Message type name
        :param str user_id: Creating user id
        :param str app_id: Creating application id
        :rtype: pika.BasicProperties

        """
        return pika.BasicProperties(content_type, content_encoding, headers,
                                    delivery_mode, priority, correlation_id,
                                    reply_to, expiration, message_id, timestamp,
                                    message_type, user_id, app_id)

    def _new_rabbitmq_connection(self):
        """Return a connection to RabbitMQ via the pika.Connection object.
        When RabbitMQ is connected, on_rabbitmq_open will be called.

        :rtype: pika.adapters.tornado_connection.TornadoConnection

        """
        return tornado_connection.TornadoConnection(self._rabbitmq_parameters,
                                                    self.on_rabbitmq_conn_open)

    def _publish_deferred_messages(self):
        """Called when pika is connected and has a channel open to publish
        any requests buffered.

        """
        global message_stack
        if not self._rabbitmq_is_closed and message_stack:
            LOGGER.info('Publishing %i deferred message(s)', len(message_stack))
            while message_stack:
                self._publish_message(*message_stack.pop())

    def _publish_message(self, exchange, routing_key, message, properties):
        """Publish the message to RabbitMQ

        :param str exchange: The exchange to publish to
        :param str routing_key: The routing key to publish with
        :param str message: The message body
        :param pika.BasicProperties: The message properties

        """
        if self._rabbitmq_is_closed or not self._rabbitmq_channel:
            LOGGER.warning('Temporarily buffering message to publish')
            self._add_to_publish_stack(exchange, routing_key,
                                       message, properties)
            return
        self._rabbitmq_channel.basic_publish(exchange, routing_key,
                                             message, properties)

    @property
    def _rabbitmq_config(self):
        """Return the RabbitMQ configuration dictionary.

        :rtype: dict

        """
        config = self.application.settings.get('rabbitmq')
        if not config:
            raise exceptions.ConfigurationException('rabbitmq')
        return config

    @property
    def _rabbitmq_channel(self):
        """Return the Pika channel from the tinman object assignment.

        :rtype: pika.channel.Channel

        """
        return getattr(self.application.attributes, self.CHANNEL, None)

    @property
    def _rabbitmq_is_closed(self):
        """Returns True if the pika connection to RabbitMQ is closed.

        :rtype: bool

        """
        global rabbitmq_connection
        return not rabbitmq_connection and not pending_rabbitmq_connection

    @property
    def _rabbitmq_parameters(self):
        """Return a pika ConnectionParameters object using the configuration
        from the configuration service. The configuration dictionary should
        match the parameters for pika.connection.ConnectionParameters and
        include an extra username and password variable.

        :rtype: pika.ConnectionParameters

        """
        kwargs = dict(self._rabbitmq_config)
        kwargs['credentials'] =  pika.PlainCredentials(kwargs['username'],
                                                       kwargs['password'])
        for key in ['username', 'password']:
            del kwargs[key]
        return pika.ConnectionParameters(**kwargs)

    def _set_rabbitmq_channel(self, channel):
        """Assign the channel object to the tinman global object.

        :param pika.channel.Channel channel: The pika channel

        """
        setattr(self.application.attributes, self.CHANNEL, channel)

    def on_rabbitmq_close(self, reply_code, reply_text):
        """Called when RabbitMQ has been connected to.

        :param int reply_code: The code for the disconnect
        :param str reply_text: The disconnect reason

        """
        global rabbitmq_connection
        LOGGER.warning('RabbitMQ has disconnected (%s): %s',
                       reply_code, reply_text)
        rabbitmq_connection = None
        self._set_rabbitmq_channel(None)
        self._connect_to_rabbitmq()

    def on_rabbitmq_conn_open(self, connection):
        """Called when RabbitMQ has been connected to.

        :param pika.connection.Connection connection: The pika connection

        """
        global pending_rabbitmq_connection, rabbitmq_connection
        LOGGER.info('RabbitMQ has connected')
        rabbitmq_connection = connection
        rabbitmq_connection.add_on_close_callback(self.on_rabbitmq_close)
        rabbitmq_connection.channel(self.on_rabbitmq_channel_open)
        pending_rabbitmq_connection = None

    def on_rabbitmq_channel_open(self, channel):
        """Called when the RabbitMQ accepts the channel open request.

        :param pika.channel.Channel channel: The channel opened with RabbitMQ

        """
        LOGGER.info('Channel %i is opened for communication with RabbitMQ',
                    channel.channel_number)
        self._set_rabbitmq_channel(channel)
        self._publish_deferred_messages()

    def prepare(self):
        """Prepare the handler, ensuring RabbitMQ is connected or start a new
        connection attempt.

        """
        super(RabbitMQRequestHandler, self).prepare()
        if self._rabbitmq_is_closed:
            self._connect_to_rabbitmq()



"""Custom Tinman Handlers add wrappers to based functionality to speed
application development.

"""
HEAD = 'HEAD'
GET = 'GET'
POST = 'POST'
DELETE = 'DELETE'
PATCH = 'PATCH'
PUT = 'PUT'
OPTIONS = 'OPTIONS'

from tinman.handlers.base import RequestHandler
from tinman.handlers.session import SessionRequestHandler



"""
Redis Session Adapter

"""
import logging
from tinman import session
import redis

LOGGER = logging.getLogger(__name__)


class RedisSessionAdapter(session.SessionAdapter):
    """Stores session data in redis. Configuration values that can be set:

        host, port, db

    If the host, port and db are not set, the

    """
    FORMAT = 'sess:%s'
    REDIS = 'session_redis'
    REDIS_HOST = 'localhost'
    REDIS_PORT = 6379
    REDIS_DB = 0

    def __init__(self, application, session_id=None, configuration=None,
                 duration=session.DEFAULT_DURATION, serializer=None):
        """Create a session adapter for the base URL specified, creating a new
        session if no session id is passed in.

        :param tinman.application.Application application: The Tinman app
        :param str session_id: The current session id if once has been started
        :param dict configuration: Session storage configuration
        :param int duration: The session duration for storage retention
        :type serializer: tinman.session.serializer.SessionSerializer

        """
        LOGGER.debug('Creating a new session adapter: %r', duration)
        super(RedisSessionAdapter, self).__init__(application, session_id,
                                                  configuration, duration,
                                                  serializer)
        self._setup_redis_client()
        LOGGER.debug('Duration: %r', self._duration)

    def delete(self):
        """Remove the session data from storage, clearing any session values"""
        yield
        self._redis.delete(self._session_key)
        self.clear()

    def save(self):
        """Store the session for later retrieval"""
        pipe = self._redis.pipeline()
        pipe.set(self._session_key, self._serialize())
        pipe.expire(self._session_key, self._duration)
        pipe.execute()
        LOGGER.debug('Session saved')

    def _load_session_data(self):
        """Extend to the load the session from storage

        :rtype: dict

        """
        data = self._redis.get(self._session_key)
        return self._deserialize(data) if data else dict()

    def _redis_client(self):
        """Return a newly constructed redis connection.

        :rtype: redis.Redis

        """
        LOGGER.debug(self._redis_config)
        return redis.Redis(**self._redis_config)

    @property
    def _redis_config(self):
        """Return the redis configuration as a dict.

        :return: dict

        """
        return {'host': self._config.get('host', self.REDIS_HOST),
                'port': self._config.get('port', self.REDIS_PORT),
                'db': self._config.get('db', self.REDIS_DB)}

    @property
    def _session_key(self):
        """Return the session key

        :rtype: str

        """
        return self.FORMAT % self.id

    def _setup_redis_client(self):
        """Setup a new redis client if it's not setup at the application
        attribute level, otherwise assign the attribute to an attribute of this
        object instance.

        """
        if self.REDIS not in self._application.attributes:
            self._redis = self._redis_client()
            self._application.attributes.add(self.REDIS, self._redis)
        else:
            self._redis = self._application.attributes.session_redis



"""The RedisRequestHandler uses tornado-redis to support Redis. It will
auto-establish a single redis connection when initializing the connection.

"""
import logging
from tornado import web

LOGGER = logging.getLogger(__name__)


class RedisRequestHandler(web.RequestHandler):
    """This request handler will connect to Redis on initialize if the
    connection is not previously set. This handler uses the redis library for
    synchronous redis use.

    """
    CONFIG_DB = 'db'
    REDIS = 'redis'
    REDIS_HOST = 'localhost'
    REDIS_PORT = 6379
    REDIS_DB = 0

    def _new_redis_client(self):
        """Create a new redis client and assign it to the application.attributes
        object for reuse later.

        """
        if 'redis' not in globals():
            import redis
        LOGGER.info('Creating new Redis instance')
        settings = self._redis_settings
        LOGGER.debug('Connecting to redis: %r', settings)
        client = redis.Redis(**settings)
        self.application.attributes.add(self.REDIS, client)

    @property
    def _redis_settings(self):
        """Return the Redis settings from configuration as a dict, defaulting
        to localhost:6379:0 if it's not set in configuration. The dict format
        is set to be passed as kwargs into the Client object.

        :rtype: dict

        """
        settings = self.application.settings.get('redis', dict())
        return {'host': settings.get('host', self.REDIS_HOST),
                'port': settings.get('port', self.REDIS_PORT),
                self.CONFIG_DB: settings.get('db', self.REDIS_DB)}

    def prepare(self):
        """Prepare RedisRequestHandler requests, ensuring that there is a
        connected tornadoredis.Client object.

        """
        super(RedisRequestHandler, self).prepare()
        if self.REDIS not in self.application.attributes:
            self._new_redis_client()

    @property
    def redis(self):
        """Return a handle to the active redis client.

        :rtype: tornadoredis.Redis

        """
        if self.REDIS not in self.application.attributes:
            self._new_redis_client()
        return self.application.attributes.redis


class AsynchronousRedisRequestHandler(RedisRequestHandler):
    """This request handler will connect to Redis on initialize if the
    connection is not previously set and uses the tornado-redis library for
    asynchronous use.

    """
    CONFIG_DB = 'selected_db'

    def _new_redis_client(self):
        """Create a new redis client and assign it to the application.attributes
        object for reuse later.

        """
        if 'tornadoredis' not in globals():
            import tornadoredis
        LOGGER.info('Creating new Redis instance')
        settings = self._redis_settings
        LOGGER.debug('Connecting to redis: %r', settings)
        client = tornadoredis.Client(**settings)
        client.connect()
        self.application.attributes.add(self.REDIS, client)



"""
Tinman Whitelist Module

"""
import ipaddr
from tornado import web
import types


def whitelisted(argument=None):
    """Decorates a method requiring that the requesting IP address is
    whitelisted. Requires a whitelist value as a list in the
    Application.settings dictionary. IP addresses can be an individual IP
    address or a subnet.

    Examples:
        ['10.0.0.0/8','192.168.1.0/24', '1.2.3.4/32']

    :param list argument: List of whitelisted ip addresses or blocks
    :raises: web.HTTPError
    :raises: ValueError
    :rtype: any

    """
    def is_whitelisted(remote_ip, whitelist):
        """Check to see if an IP address is whitelisted.

        :param str ip_address: The IP address to check
        :param list whitelist: The whitelist to check against
        :rtype: bool

        """
        # Convert the ip into a long int version of the ip address
        user_ip = ipaddr.IPv4Address(remote_ip)

        # Loop through the ranges in the whitelist and check
        if any([user_ip in ipaddr.IPv4Network(entry) for entry in whitelist]):
            return True

        return False

    # If the argument is a function then there were no parameters
    if type(argument) is types.FunctionType:

        def wrapper(self, *args, **kwargs):
            """Check the whitelist against our application.settings dictionary
            whitelist key.

            :rtype: any
            :raises: web.HTTPError

            """
            # Validate we have a configured whitelist
            if 'whitelist' not in self.application.settings:
                raise ValueError('whitelist not found in Application.settings')

            # If the IP address is whitelisted, call the wrapped function
            if is_whitelisted(self.request.remote_ip,
                              self.application.settings['whitelist']):

                # Call the original function, IP is whitelisted
                return argument(self, *args, **kwargs)

            # The ip address was not in the whitelist
            raise web.HTTPError(403)

        # Return the wrapper method
        return wrapper

    # They passed in string or list?
    else:

        # Convert a single ip address to a list
        if isinstance(argument, str):
            argument = [argument]

        # Make sure it's a list
        elif not isinstance(argument, list):
            raise ValueError('whitelisted requires no parameters or '
                             'a string or list')

        def argument_wrapper(method):
            """Wrapper for a method passing in the IP addresses that constitute
            the whitelist.

            :param method method: The method being wrapped
            :rtype: any
            :raises: web.HTTPError

            """
            def validate(self, *args, **kwargs):
                """
                Validate the ip address agross the list of ip addresses
                passed in as a list
                """
                if is_whitelisted(self.request.remote_ip, argument):

                    # Call the original function, IP is whitelisted
                    return method(self, *args, **kwargs)

                # The ip address was not in the whitelist
                raise web.HTTPError(403)

            # Return the validate method
            return validate

        # Return the wrapper method
        return argument_wrapper



"""
Tinman Test Application

"""
import logging
from tornado import web

from tinman.handlers import SessionRequestHandler
from tinman import __version__

LOGGER = logging.getLogger(__name__)


CONFIG = {'Application': {'debug': True,
                          'xsrf_cookies': False},
          'HTTPServer': {'no_keep_alive': False,
                         'ports': [8000],
                         'xheaders': False},
          'Logging': {'loggers': {'tinman': {'propagate': True,
                                            'level': 'DEBUG'}},
                      'formatters': {'verbose': ('%(levelname) -10s %(asctime)s'
                                                 ' %(name) -30s %(funcName) '
                                                 '-25s: %(message)s')},
                      'filters': {'tinman': 'tinman'},
                      'handlers': {'console': {'formatter': 'verbose',
                                               'filters': ['tinman'],
                                               'debug_only': True,
                                               'class': 'logging.StreamHandler',
                                               'level': 'DEBUG'},
                                   'file': {'delay': False,
                                            'mode': 'a',
                                            'encoding': 'UTF-8',
                                            'formatter': 'verbose',
                                            'filters': ['tinman'],
                                            'class': 'logging.FileHandler',
                                            'filename': '/tmp/tinman.log'}}},
          'Routes': [("/", "tinman.test.DefaultHandler")]}


class Handler(SessionRequestHandler):

    @web.asynchronous
    def get(self, *args, **kwargs):
        """Example HTTP Get response method.

        :param args: positional args
        :param kwargs: keyword args

        """
        self.session.username = 'gmr'

        session = self.session.as_dict()
        if session['last_request_at']:
            session['last_request_at'] = session['last_request_at'].isoformat()

        # Send a JSON string for our test
        self.write({'message': 'Hello World',
                    'request': {'method': self.request.method,
                                'protocol': self.request.protocol,
                                'path': self.request.path,
                                'query': self.request.query,
                                'remote_ip': self.request.remote_ip,
                                'version': self.request.version},
                    'session': session,
                    'tinman': {'version':  __version__}})
        self.finish()



"""
Configuration Constants

"""
APPLICATION = 'Application'
DAEMON = 'Daemon'
HTTP_SERVER = 'HTTPServer'
LOGGING = 'Logging'
ROUTES = 'Routes'

AUTOMATIC = 'automatic'
BASE = 'base'
BASE_VARIABLE = '{{base}}'
CERT_REQS = 'cert_reqs'
DEBUG = 'debug'
DEFAULT_LOCALE = 'default_locale'
FILE = 'file'
LOG_FUNCTION = 'log_function'
NEWRELIC = 'newrelic_ini'
NO_KEEP_ALIVE = 'no_keep_alive'
NONE = 'none'
OPTIONAL = 'optional'
PROCESSES = 'processes'
PATHS = 'paths'
PORTS = 'ports'
RABBITMQ = 'rabbitmq'
REDIS = 'redis'
REQUIRED = 'required'
SSL_OPTIONS = 'ssl_options'
STATIC = 'static'
TEMPLATES = 'templates'
TRANSFORMS = 'transforms'
TRANSLATIONS = 'translations'
UI_MODULES = 'ui_modules'
VERSION = 'version'
XHEADERS = 'xheaders'



"""
A tornado.web.RequestHandler decorator that provides HTTP Basic Authentication. 

The decorator takes two arguments: 

    1. realm: the realm that's typically presented to the user during a
    challenge request for authentication.

    2. validate_callback: A callable that's used to validate the credentials.
    The callable will receive the username and password provided by the end
    user in a challenge.

Example usage (also see helloworld_basic.py in the examples): 

    # define the validation callback.
    def validate(uname, passwd):
        creds = {'auth_username': 'jonesy', 'auth_password': 'foobar'}
        if uname == creds['auth_username'] and passwd == creds['auth_password']:
            return True
        else:
            return False

    # now define the RequestHandler, using the decorator.
    @require_basic_auth('AuthRealm', validate)
    class MainHandler(tornado.web.RequestHandler):
        def get(self):
            self.write("Hello, world - Tornado %s" % tornado.version)

""" 

import base64

def require_basic_auth(realm, validate_callback, do_wrap=True):
    def require_basic_auth_decorator(handler_class):
        def wrap_execute(handler_execute):
            def require_basic_auth(handler, kwargs):
                def create_auth_header():
                    print("Creating auth header")
                    handler.set_status(401)
                    handler.set_header('WWW-Authenticate', 'Basic realm=%s' % realm)
                    handler._transforms = []
                    handler.finish()

                auth_header = handler.request.headers.get('Authorization')
                if auth_header is None or not auth_header.startswith('Basic '):
                    create_auth_header()
                else:
                    auth_decoded = base64.decodestring(auth_header[6:])
                    basicauth_user, basicauth_pass = auth_decoded.split(':', 2)
                    if validate_callback(basicauth_user, basicauth_pass):
                        return True
                    else:
                        create_auth_header()
            def _execute(self, transforms, *args, **kwargs):
                if not require_basic_auth(self, kwargs):
                    return False
                return handler_execute(self, transforms, *args, **kwargs)
            return _execute

        if do_wrap:
            handler_class._execute = wrap_execute(handler_class._execute)
        return handler_class
    return require_basic_auth_decorator



"""
process.py

"""
from helper import config as helper_config
from tornado import httpserver
from tornado import ioloop
import logging
import multiprocessing
import signal
import socket
import ssl
from tornado import version as tornado_version

from tinman import application
from tinman import config
from tinman import exceptions

LOGGER = logging.getLogger(__name__)


class Process(multiprocessing.Process):
    """The process holding the HTTPServer and Application"""
    CERT_REQUIREMENTS = {config.NONE: ssl.CERT_NONE,
                         config.OPTIONAL: ssl.CERT_OPTIONAL,
                         config.REQUIRED: ssl.CERT_REQUIRED}
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        """Create a new instance of Process

        """
        super(Process, self).__init__(group, target, name, args, kwargs)

        # Passed in values
        self.namespace = kwargs['namespace']
        self.port = kwargs['port']

        # Internal attributes holding instance information
        self.app = None
        self.http_server = None
        self.request_counters = dict()

        # Re-setup logging in the new process
        self.logging_config = None

        # If newrelic is passed, use it
        if self.newrelic_ini_path:
            self.setup_newrelic()

    def create_application(self):
        """Create and return a new instance of tinman.application.Application"""
        return application.Application(self.settings,
                                       self.namespace.routes,
                                       self.port)

    def create_http_server(self):
        """Setup the HTTPServer

        :rtype: tornado.httpserver.HTTPServer

        """
        return self.start_http_server(self.port, self.http_config)

    @property
    def http_config(self):
        """Return a dictionary of HTTPServer arguments using the default values
        as specified in the HTTPServer class docstrings if no values are
        specified.

        :param dict config: The HTTPServer specific section of the config
        :rtype: dict

        """
        return {config.NO_KEEP_ALIVE:
                    self.namespace.server.get(config.NO_KEEP_ALIVE, False),
                config.SSL_OPTIONS: self.ssl_options,
                config.XHEADERS: self.namespace.server.get(config.XHEADERS,
                                                           False)}

    def on_sigabrt(self, signal_unused, frame_unused):
        """Stop the HTTP Server and IO Loop, shutting down the process

        :param int signal_unused: Unused signal number
        :param frame frame_unused: Unused frame the signal was caught in

        """
        LOGGER.info('Stopping HTTP Server and IOLoop')
        self.http_server.stop()
        self.ioloop.stop()

    def on_sighup(self, signal_unused, frame_unused):
        """Reload the configuration

        :param int signal_unused: Unused signal number
        :param frame frame_unused: Unused frame the signal was caught in

        """
        # Update HTTP configuration
        for setting in self.http_config:
            if getattr(self.http_server, setting) != self.http_config[setting]:
                LOGGER.debug('Changing HTTPServer %s setting', setting)
                setattr(self.http_server, setting, self.http_config[setting])

        # Update Application Settings
        for setting in self.settings:
            if self.app.settings[setting] != self.settings[setting]:
                LOGGER.debug('Changing Application %s setting', setting)
                self.app.settings[setting] = self.settings[setting]

        # Update the routes
        self.app.handlers = []
        self.app.named_handlers = {}
        routes = self.namespace.config.get(config.ROUTES)
        self.app.add_handlers(".*$", self.app.prepare_routes(routes))

        LOGGER.info('Configuration reloaded')

    def run(self):
        """Called when the process has started

        :param int port: The HTTP Server port

        """
        LOGGER.debug('Initializing process')

        # Setup logging
        self.logging_config = self.setup_logging()

        # Register the signal handlers
        self.setup_signal_handlers()

        # Create the application instance
        try:
            self.app = self.create_application()
        except exceptions.NoRoutesException:
            return

        # Create the HTTPServer
        self.http_server = self.create_http_server()

        # Hold on to the IOLoop in case it's needed for responding to signals
        self.ioloop = ioloop.IOLoop.instance()

        # Start the IOLoop, blocking until it is stopped
        try:
            self.ioloop.start()
        except KeyboardInterrupt:
            pass

    @property
    def settings(self):
        """Return the Application configuration

        :rtype: dict

        """
        return dict(self.namespace.config)

    def setup_logging(self):
        return helper_config.LoggingConfig(self.namespace.logging)

    @property
    def newrelic_ini_path(self):
        return self.namespace.config.get(config.NEWRELIC)

    def setup_newrelic(self):
        """Setup the NewRelic python agent"""
        import newrelic.agent
        newrelic.agent.initialize(self.newrelic_ini_path)

    def setup_signal_handlers(self):
        """Called when a child process is spawned to register the signal
        handlers

        """
        LOGGER.debug('Registering signal handlers')
        signal.signal(signal.SIGABRT, self.on_sigabrt)

    @property
    def ssl_options(self):
        """Check the config to see if SSL configuration options have been passed
        and replace none, option, and required with the correct values in
        the certreqs attribute if it is specified.

        :rtype: dict

        """
        opts = self.namespace.server.get(config.SSL_OPTIONS) or dict()
        if config.CERT_REQS in opts:
            opts[config.CERT_REQS] = \
                self.CERT_REQUIREMENTS[opts[config.CERT_REQS]]
        return opts or None

    def start_http_server(self, port, args):
        """Start the HTTPServer

        :param int port: The port to run the HTTPServer on
        :param dict args: Dictionary of arguments for HTTPServer
        :rtype: tornado.httpserver.HTTPServer

        """
        # Start the HTTP Server
        LOGGER.info("Starting Tornado v%s HTTPServer on port %i Args: %r",
                    tornado_version, port, args)
        http_server = httpserver.HTTPServer(self.app, **args)
        http_server.bind(port, family=socket.AF_INET)
        http_server.start(1)
        return http_server



"""
Tinman provides a base RequestHandler extending Tornado's web.RequestHandler
adding additional behaviors across all RequestHandlers that use it.

"""
import json
from tornado import web


class RequestHandler(web.RequestHandler):
    """A base RequestHandler that adds the following functionality:

    - If sending a dict, checks the user-agent string for curl and sends an
      indented, sorted human-readable JSON snippet.
    - Toggles the ensure_ascii flag in json.dumps
    - Overrides the default behavior for unimplemented methods to instead set
    the status and look to the allow object attribute for methods that can be
    allowed. This is useful for when using NewRelic since the newrelic agent
    will catch the normal exceptions thrown as errors and trigger false alerts.

    To use, do something like::

        from tinman import handlers

        class Handler(handlers.RequestHandler):

            allow = [handlers.GET, handlers.POST]

            def get(self, *args, **kwargs):
                self.write({'foo': 'bar'})

            def post(self, *args, **kwargs):
                self.write({'message': 'Saved'})

    """
    allow = []

    def _method_not_allowed(self):
        self.set_header('Allow', ', '.join(self.allow))
        self.set_status(405, 'Method Not Allowed')

    def head(self, *args, **kwargs):
        """Implement the HTTP HEAD method

        :param list args: Positional arguments
        :param dict kwargs: Keyword arguments

        """
        self._method_not_allowed()

    def get(self, *args, **kwargs):
        """Implement the HTTP GET method

        :param list args: Positional arguments
        :param dict kwargs: Keyword arguments

        """
        self._method_not_allowed()

    def post(self, *args, **kwargs):
        """Implement the HTTP POST method

        :param list args: Positional arguments
        :param dict kwargs: Keyword arguments

        """
        self._method_not_allowed()

    def delete(self, *args, **kwargs):
        """Implement the HTTP DELETE method

        :param list args: Positional arguments
        :param dict kwargs: Keyword arguments

        """
        self._method_not_allowed()

    def patch(self, *args, **kwargs):
        """Implement the HTTP PATCH method

        :param list args: Positional arguments
        :param dict kwargs: Keyword arguments

        """
        self._method_not_allowed()

    def put(self, *args, **kwargs):
        """Implement the HTTP PUT method

        :param list args: Positional arguments
        :param dict kwargs: Keyword arguments

        """
        self._method_not_allowed()

    def options(self, *args, **kwargs):
        """Implement the HTTP OPTIONS method

        :param list args: Positional arguments
        :param dict kwargs: Keyword arguments

        """
        self._method_not_allowed()

    def write(self, chunk):
        """Writes the given chunk to the output buffer.

        To write the output to the network, use the flush() method below.

        If the given chunk is a dictionary, we write it as JSON and set
        the Content-Type of the response to be ``application/json``.
        (if you want to send JSON as a different ``Content-Type``, call
        set_header *after* calling write()).

        :param mixed chunk: The string or dict to write to the client

        """
        if self._finished:
            raise RuntimeError("Cannot write() after finish().  May be caused "
                               "by using async operations without the "
                               "@asynchronous decorator.")
        if isinstance(chunk, dict):
            options = {'ensure_ascii': False}
            if 'curl' in self.request.headers.get('user-agent'):
                options['indent'] = 2
                options['sort_keys'] = True
            chunk = json.dumps(chunk, **options).replace("</", "<\\/") + '\n'
            self.set_header("Content-Type", "application/json; charset=UTF-8")
        self._write_buffer.append(web.utf8(chunk))



"""
See an example that uses basic auth with an LDAP 
backend in examples/helloworld_basic_ldap.py

"""
import ldap
import logging

# where to start the search for users
LDAP_SEARCH_BASE = 'ou=People,dc=yourdomain,dc=com'

# the server to auth against
LDAP_URL = 'ldap://ldap.yourdomain.com'

# The attribute we try to match the username against.
LDAP_UNAME_ATTR = 'uid'

# The attribute we need to retrieve in order to perform a bind.
LDAP_BIND_ATTR = 'dn'

# Whether to use LDAPv3. Highly recommended.
LDAP_VERSION_3 = True

def auth_user_ldap(uname, pwd):
    """
    Attempts to bind using the uname/pwd combo passed in.
    If that works, returns true. Otherwise returns false.

    """
    if not uname or not pwd:
        logging.error("Username or password not supplied")
        return False

    ld = ldap.initialize(LDAP_URL)
    if LDAP_VERSION_3:
        ld.set_option(ldap.VERSION3, 1)
    ld.start_tls_s()
    udn = ld.search_s(LDAP_SEARCH_BASE, ldap.SCOPE_ONELEVEL,
                           '(%s=%s)' % (LDAP_UNAME_ATTR,uname), [LDAP_BIND_ATTR])
    if udn:
        try:
            bindres = ld.simple_bind_s(udn[0][0], pwd)
        except ldap.INVALID_CREDENTIALS, ldap.UNWILLING_TO_PERFORM:
            logging.error("Invalid or incomplete credentials for %s", uname)
            return False
        except Exception as out:
            logging.error("Auth attempt for %s had an unexpected error: %s",
                         uname, out)
            return False
        else:
            return True
    else:
        logging.error("No user by that name")
        return False





"""
The CouchDB template loader allows for Tornado templates to be stored in CouchDB
and retrieved on demand and supports all of the syntax of including and
extending templates that you'd expect in any other template loader.

"""
import json
import logging
from tornado import escape
from tornado import httpclient
from tornado import template

LOGGER = logging.getLogger(__name__)


class CouchDBLoader(template.BaseLoader):
    """Extends the tornado.template.Loader allowing for templates to be loaded
    out of CouchDB.

    Templates in CouchDB should have have an _id matching the value of the name
    that is passed into load. _id's may have /'s in them. The template itself
    should be in the template node of the JSON document in CouchDB.

    """
    def __init__(self, base_url, **kwargs):
        """Creates a template loader.

        :param str base_url: The base URL for the CouchDB server

        """
        super(CouchDBLoader, self).__init__('/', **kwargs)
        self._base_url = base_url.rstrip('/')
        LOGGER.info('Initialized with base URL of %s', self._base_url)
        self._http_client = httpclient.HTTPClient()

    def load(self, name, parent_path=None):
        """Loads a template.

        :param str name: The template name
        :param str parent_path: The optional path for a parent document
        :rtype: tornado.template.Template

        """
        if name not in self.templates:
            self.templates[name] = self._create_template(name)
        return self.templates[name]

    def _create_template(self, name):
        """Create an instance of a tornado.template.Template object for the
        given template name.

        :param str name: The name/path to the template
        :rtype: tornado.template.Template

        """
        url = '%s/%s' % (self._base_url, escape.url_escape(name))
        LOGGER.debug('Making HTTP GET request to %s', url)
        response = self._http_client.fetch(url)
        data = json.loads(response.body, ensure_ascii=False)
        return template.Template(data['template'], name=name, loader=self)



"""
@TODO see if we can move these functions to a more appropriate spot

"""
import importlib
import os
import sys
from socket import gethostname


def application_name():
    """Returns the currently running application name

    :rtype: str

    """
    return os.path.split(sys.argv[0])[1]


def hostname():
    """Returns the hostname for the machine we're running on

    :rtype: str

    """
    return gethostname().split(".")[0]


def import_namespaced_class(path):
    """Pass in a string in the format of foo.Bar, foo.bar.Baz, foo.bar.baz.Qux
    and it will return a handle to the class

    :param str path: The object path
    :rtype: class

    """
    parts = path.split('.')
    return getattr(importlib.import_module('.'.join(parts[0:-1])), parts[-1])



"""The SessionRequestHandler implements a base level of session handling. See
the tinman.session package to see what types of session adapters exist.

"""
import datetime
import logging

LOGGER = logging.getLogger(__name__)

from tinman import session
from tinman.handlers import base


class SessionRequestHandler(base.RequestHandler):
    """A RequestHandler that implements user sessions.  Configuration in the
    application settings is as follows:

    Application:
      session:
        adapter:
          class: FileSessionAdapter
          configuration: SessionAdapter specific configuration
        cookie:
          name: session
        duration: 3600

    All options other than the adapter class are optional.

    """
    ADAPTER = 'adapter'
    COOKIE = 'cookie'
    DURATION = 'duration'
    NAME = 'name'

    COOKIE_NAME = 'session'
    DEFAULT_DURATION = 3600

    def on_finish(self):
        """Called by Tornado when the request is done. Update the session data
        and remove the session object.

        """
        super(SessionRequestHandler, self).on_finish()
        self.session.last_request_at = datetime.datetime.now().strftime('%s')
        self.session.last_request_uri = self.request.uri
        self.session.save()
        del self.session

    def prepare(self):
        """Prepare the session, setting up the session object and loading in
        the values, assigning the IP address to the session if it's an new one.

        """
        super(SessionRequestHandler, self).prepare()
        self.session = self._get_session_object()
        self.session.load()
        if not self.session.ip_address:
            self.session.ip_address = self.request.remote_ip
        self._last_values()
        self._set_session_cookie()
        LOGGER.debug('Session ID: %s', self.session.id)

    def _last_values(self):
        """Always carry last_request_uri and last_request_at even if the last_*
        values are null.

        """
        if not self.session.last_request_uri:
            self.session.last_request_uri = None
        lra = self.session.last_request_at
        if lra is None:
            self.session.last_request_at = None
        else:
            lra = float(lra)
            self.session.last_request_at = datetime.datetime.fromtimestamp(lra)

    def _clear_session(self):
        """Clear the user's sessions, resetting the cookies and removing the
        data from redis.

        """
        LOGGER.info('Clearing session')
        self.session.delete()
        self.clear_cookie(self._session_cookie_name)

    @property
    def _cookie_session_id(self):
        """Gets the session id from the session cookie.

        :rtype: str

        """
        return self.get_secure_cookie(self._session_cookie_name, None)

    def _get_session_object(self):
        """Return an instance of the session object for the current session.
        If there is no pre-existing session, the session object will be created
        with a new session id.

        :rtype: tinman.session.SessionAdapter

        """
        return session.get_session_adapter(self.application,
                                           self._cookie_session_id,
                                           self._adapter_settings,
                                           self._duration)

    @property
    def _adapter_settings(self):
        """Return the session adapter settings

        :rtype: dict

        """
        return self._settings.get(self.ADAPTER) or dict()

    @property
    def _cookie_expiration(self):
        """Return the expiration timestamp for the session cookie.

        :rtype: datetime

        """
        value = (datetime.datetime.utcnow() +
                 datetime.timedelta(seconds=self._duration))
        LOGGER.debug('Cookie expires: %s', value.isoformat())
        return value

    @property
    def _session_cookie_name(self):
        """Return the session cookie name, defaulting to the class default

        :rtype: str

        """
        return self._cookie_settings.get(self.NAME, self.COOKIE_NAME)

    @property
    def _cookie_settings(self):
        """Return the cookie specific session settings

        :rtype: dict

        """
        return self._settings.get(self.COOKIE)

    @property
    def _duration(self):
        """Return the session duration in seconds from the configuration,
        defaulting to the class default.

        :rtype: int

        """
        return self._settings.get(self.DURATION, self.DEFAULT_DURATION)

    @property
    def _settings(self):
        """Return the session management settings.

        :rtype: dict

        """
        return self.application.settings.get('session') or dict()

    def _set_session_cookie(self):
        """Set the session data cookie."""
        LOGGER.debug('Setting session cookie for %s', self.session.id)
        self.set_secure_cookie(name=self._session_cookie_name,
                               value=self.session.id,
                               expires=self._cookie_expiration)



"""
Tinman Exceptions

"""

class ConfigurationException(Exception):
    def __repr__(self):
        return 'Configuration for %s is missing or invalid' % self.args[0]


class NoRoutesException(Exception):
    def __repr__(self):
        return 'No routes could be configured'


"""
Main Tinman Application Class

"""
import logging
import sys
from tornado import web

from tinman import config
from tinman import exceptions
from tinman import utils
from tinman import __version__

LOGGER = logging.getLogger(__name__)

STATIC_PATH = 'static_path'
TEMPLATE_PATH = 'template_path'


class Application(web.Application):
    """Application extends web.Application and handles all sorts of things
    for you that you'd have to handle yourself.

    """
    def __init__(self, settings, routes, port):
        """Create a new Application instance with the specified Routes and
        settings.

        :param dict settings: Application settings
        :param list routes: A list of route tuples
        :param int port: The port number for the HTTP server

        """
        self.attributes = Attributes()
        self.host = utils.gethostname()
        self.port = port
        self._config = settings or dict()
        self._insert_base_path()
        self._prepare_paths()
        self._prepare_static_path()
        self._prepare_template_path()
        self._prepare_transforms()
        self._prepare_translations()
        self._prepare_uimodules()
        self._prepare_version()
        self.prepare_routes(routes)
        if not routes:
            LOGGER.critical('Did not add any routes, will exit')
            raise exceptions.NoRoutesException()

        # Get the routes and initialize the tornado.web.Application instance
        super(Application, self).__init__(routes, **self._config)

    def log_request(self, handler):
        """Writes a completed HTTP request to the logs.

        By default writes to the tinman.application LOGGER.  To change
        this behavior either subclass Application and override this method,
        or pass a function in the application settings dictionary as
        'log_function'.

        :param tornado.web.RequestHandler handler: The request handler

        """
        if config.LOG_FUNCTION in self.settings:
            self.settings[config.LOG_FUNCTION](handler)
            return
        if handler.get_status() < 400:
            log_method = LOGGER.info
        elif handler.get_status() < 500:
            log_method = LOGGER.warning
        else:
            log_method = LOGGER.exception
        request_time = 1000.0 * handler.request.request_time()
        log_method("%d %s %.2fms", handler.get_status(),
                   handler._request_summary(), request_time)

    @property
    def paths(self):
        """Return the path configuration

        :rtype: dict

        """
        return self._config.get(config.PATHS, dict())

    def prepare_routes(self, routes):
        """Prepare the routes by iterating through the list of tuples & calling
        prepare route on them.

        :param routes: Routes to prepare
        :type routes: list
        :rtype: list
        :raises: ValueError

        """
        if not isinstance(routes, list):
            raise ValueError('Routes parameter must be a list of tuples')
        prepared_routes = list()
        for parts in routes:
            route = self._prepare_route(parts)
            if route:
                LOGGER.info('Appending handler: %r', route)
                prepared_routes.append(route)
        return prepared_routes

    def _import_class(self, class_path):
        """Try and import the specified namespaced class.

        :param str class_path: The full path to the class (foo.bar.Baz)
        :rtype: class

        """
        LOGGER.debug('Importing %s', class_path)
        try:
            return utils.import_namespaced_class(class_path)
        except ImportError as error:
            LOGGER.critical('Could not import %s: %s', class_path, error)
            return None

    def _import_module(self, module_path):
        """Dynamically import a module returning a handle to it.

        :param str module_path: The module path
        :rtype: module

        """
        LOGGER.debug('Importing %s', module_path)
        try:
            return __import__(module_path)
        except ImportError as error:
            LOGGER.critical('Could not import %s: %s', module_path, error)
            return None

    def _insert_base_path(self):
        """If the "base" path is set in the paths section of the config, insert
        it into the python path.

        """
        if config.BASE in self.paths:
            sys.path.insert(0, self.paths[config.BASE])

    def _prepare_paths(self):
        """Set the value of {{base}} in paths if the base path is set in the
        configuration.

        :raises: ValueError

        """
        if config.BASE in self.paths:
            for path in [path for path in self.paths if path != config.BASE]:
                if config.BASE_VARIABLE in self.paths[path]:
                    self.paths[path] = \
                        self.paths[path].replace(config.BASE_VARIABLE,
                                                 self.paths[config.BASE])
        LOGGER.debug('Prepared paths: %r', self.paths)

    def _prepare_route(self, attrs):
        """Take a given inbound list for a route and parse it creating the
        route and importing the class it belongs to.

        :param list attrs: Route attributes
        :rtype: list

        """
        if type(attrs) not in (list, tuple):
            LOGGER.error('Invalid route, must be a list or tuple: %r', attrs)
            return

        # By default there are not any extra kwargs
        kwargs = None

        # If there is a regex based route, set it up with a raw string
        if attrs[0] == 're':
            route = r'%s' % attrs[1]
            classpath = attrs[2]
            if len(attrs) == 4:
                kwargs = attrs[3]
        else:
            route = r'%s' % attrs[0]
            classpath = attrs[1]
            if len(attrs) == 3:
                kwargs = attrs[2]

        LOGGER.debug('Initializing route: %s with %s', route, classpath)
        try:
            handler = self._import_class(classpath)
        except ImportError as error:
            LOGGER.error('Class import error for %s: %r', classpath, error)
            return None

        # Setup the prepared route, adding kwargs if there are any
        prepared_route = [route, handler]
        if kwargs:
            prepared_route.append(kwargs)

        # Return the prepared route as a tuple
        return tuple(prepared_route)

    def _prepare_static_path(self):
        LOGGER.info('%s in %r: %s', config.STATIC, self.paths,
                    config.STATIC in self.paths)
        if config.STATIC in self.paths:
            LOGGER.info('Setting static path to %s', self.paths[config.STATIC])
            self._config[STATIC_PATH] = self.paths[config.STATIC]

    def _prepare_template_path(self):
        LOGGER.info('%s in %r: %s', config.TEMPLATES, self.paths,
                    config.TEMPLATES in self.paths)
        if config.TEMPLATES in self.paths:
            LOGGER.info('Setting template path to %s',
                        self.paths[config.TEMPLATES])
            self._config[TEMPLATE_PATH] = self.paths[config.TEMPLATES]

    def _prepare_transforms(self):
        """Prepare the list of transforming objects"""
        if config.TRANSFORMS in self._config:
            LOGGER.info('Preparing %i transform class(es) for import',
                        len(self._config[config.TRANSFORMS]))
            for transform in [self._import_module(transform) for transform in
                              self._config[config.TRANSFORMS]]:
                LOGGER.debug('Adding transform: %r', transform)
                self.add_transform(transform)

    def _prepare_translations(self):
        """Load in translations if they are set, and add the default locale as
        well.

        """
        if config.TRANSLATIONS in self.paths:
            LOGGER.info('Loading translations from %s',
                        self.paths[config.TRANSLATIONS])
            from tornado import locale
            locale.load_translations(self.paths[config.TRANSLATIONS])
            if config.DEFAULT_LOCALE in self._config:
                LOGGER.info('Setting default locale to %s',
                            self._config[config.DEFAULT_LOCALE])
                locale.set_default_locale(self._config[config.DEFAULT_LOCALE])

    def _prepare_uimodule(self):
        self._config[config.UI_MODULES] = \
            self._import_module(self._config[config.UI_MODULES])

    def _prepare_uimodule_dict(self):
        for key, value in self._config[config.UI_MODULES].items():
            self._config[config.UI_MODULES][key] = self._import_module(value)

    def _prepare_uimodule_list(self):
        for offset, value in enumerate(self._config[config.UI_MODULES]):
            self._config[config.UI_MODULES][offset] = self._import_module(value)

    def _prepare_uimodules(self):
        """Prepare the UI Modules object, handling the three cases that Tornado
        supports for the ui_modules configuration: a single module, a mapping
        of modules in a dictionary or a list of modules.

        """
        if config.UI_MODULES in self._config:
            if isinstance(self._config[config.UI_MODULES], str):
                self._prepare_uimodule()
            elif isinstance(self._config[config.UI_MODULES], dict()):
                self._prepare_uimodule_dict()
            elif isinstance(self._config[config.UI_MODULES], list):
                self._prepare_uimodule_list()
            else:
                LOGGER.critical('Unknown format for %s configuration: %s',
                                config.UI_MODULES,
                                type(self._config[config.UI_MODULES]))

    def _prepare_version(self):
        """Setup the application version"""
        if config.VERSION not in self._config:
            self._config[config.VERSION] = __version__


class Attributes(object):
    """A base object to hang attributes off of for application level scope that
    can be used across connections.

    """
    ATTRIBUTES = '_attributes'

    def __init__(self):
        """Create a new instance of the Attributes class"""
        self._attributes = dict()

    def __contains__(self, item):
        """Check to see if an attribute is set on the object.

        :param str item: The attribute name
        :rtype: bool

        """
        return item in self.__dict__[self.ATTRIBUTES].keys()

    def __delattr__(self, item):
        """Delete an attribute from the object.

        :param str item: The attribute name
        :raises: AttributeError

        """
        if item == self.ATTRIBUTES:
            raise AttributeError('Can not delete %s', item)
        if item not in self.__dict__[self.ATTRIBUTES]:
            raise AttributeError('%s is not set' % item)
        del self.__dict__[self.ATTRIBUTES][item]

    def __getattr__(self, item):
        """Get an attribute from the class.

        :param str item: The attribute name
        :rtype: any

        """
        if item == self.ATTRIBUTES:
            return self.__dict__[item]
        return self.__dict__[self.ATTRIBUTES].get(item)

    def __iter__(self):
        """Iterate through the keys in the data dictionary.

        :rtype: list

        """
        return iter(self.__dict__[self.ATTRIBUTES])

    def __len__(self):
        """Return the length of the data dictionary.

        :rtype: int

        """
        return len(self.__dict__[self.ATTRIBUTES])

    def __repr__(self):
        """Return the representation of the class as a string.

        :rtype: str

        """
        return '<%s(%r)>' % (self.__class__.__name__,
                             self.__dict__[self.ATTRIBUTES])

    def __setattr__(self, item, value):
        """Set an attribute on the object.

        :param str item: The attribute name
        :param any value: The attribute value

        """
        if item == self.ATTRIBUTES:
            self.__dict__[item] = value
        else:
            self.__dict__[self.ATTRIBUTES][item] = value

    def add(self, item, value):
        """Add an attribute value to our object instance.

        :param str item: Application attribute name
        :param any value: Value to associate with the attribute
        :raises: AttributeError

        """
        if item in self.__dict__[self.ATTRIBUTES].keys():
            raise AttributeError('%s already exists' % item)
        setattr(self, item, value)

    def remove(self, item):
        """Remove an attribute value to our object instance.

        :param str item: Application attribute name
        :raises: AttributeError

        """
        if item not in self.__dict__[self.ATTRIBUTES].keys():
            raise AttributeError('%s does not exist' % item)
        delattr(self, item)

    def set(self, item, value):
        """Set an attribute value to our object instance.

        :param str item: Application attribute name
        :param any value: Value to associate with the attribute
        :raises: AttributeError

        """
        setattr(self, item, value)



"""Create a new tinman/tornado project including setting up the setup.py file,
initial directory structure and virtual environment, if desired.

"""
import argparse
import logging
import os
import uuid


DESCRIPTION = ('A tool to create a new tinman project, including the directory '
               'structure, setup.py file and skeleton configuration')
LOGGER = logging.getLogger(__name__)

BOOTSTRAP_URL = 'http://getbootstrap.com/bs-v3.0.0-rc1-dist.zip'

from tinman import __version__


class Project(object):

    DEFAULT_MODE = 0x755
    DIRECTORIES = ['etc', 'logs', 'pid',
                   'source', 'source/less', 'source/js',
                   'static', 'static/css', 'static/img', 'static/js',
                   'templates',
                   'tests']

    def __init__(self):
        self._parser = self._create_argument_parser()

    def _add_base_arguments(self, parser):
        """Add the base arguments to the argument parser.

        :param argparse.ArgumentParser parser: The parser to add arguments to

        """
        parser.add_argument('--version',
                            help='show the version number and exit',
                            action='version',
                            version='%(prog)s ' + __version__)

    def _add_required_arguments(self, parser):
        """Add the required arguments to the argument parser.

        :param argparse.ArgumentParser parser: The parser to add arguments to

        """
        parser.add_argument('project',
                            metavar='PROJECT',
                            help='The project to create')

    def add_optional_arguments(self, parser):
        pass

    def _create_argument_parser(self):
        """Create and return the argument parser with all of the arguments
        and configuration ready to go.

        :rtype: argparse.ArgumentParser

        """
        parser = self._new_argument_parser()
        self._add_base_arguments(parser)
        self._add_required_arguments(parser)
        return parser

    def _create_base_directory(self):
        os.mkdir(self._arguments.project, self.DEFAULT_MODE)

    def _create_directories(self):
        self._create_base_directory()
        self._create_subdirectory(self._arguments.project)
        for directory in self.DIRECTORIES:
            self._create_subdirectory(directory)

    def _create_package_init(self):
        with open('%s/%s/__init__.py' %
                  (self._arguments.project,
                   self._arguments.project), 'w') as init:
            init.write('')

    def _create_config(self):

        template = """
%YAML 1.2
---
Application:
  cookie_secret: {{uuid}}
  debug: true
  stats_port: 9090
  xsrf_cookies: false
  wake_interval: 60

Daemon:
    pidfile: pid/{{app_name}}.pid

HTTPServer:
  no_keep_alive: false
  ports: [8000]
  xheaders: true

Logging:
  version: 1

  formatters:
    verbose:
      format: '%(levelname) -10s %(asctime)s %(process)-6d %(processName) -20s %(name) -20s %(funcName) -25s: %(message)s'
      datefmt: '%Y-%m-%d %H:%M:%S'
    syslog:
      format: '%(levelname)s <PID %(process)d:%(processName)s> %(name)s.%(funcName)s(): %(message)s'
    tornado:
      '()' : 'tornado.log.LogFormatter'
      color: True

  handlers:
    console:
      class: logging.StreamHandler
      debug_only: True
      formatter: verbose

    error:
      filename: logs/error.log
      class: logging.handlers.RotatingFileHandler
      maxBytes: 104857600
      backupCount: 6
      formatter: verbose

    file:
      filename: logs/request.log
      class: logging.handlers.RotatingFileHandler
      maxBytes: 104857600
      backupCount: 6
      formatter: verbose

  loggers:

    clihelper:
      level: WARNING
      propagate: true
      handlers: [console, file, error]

    tinman:
      level: INFO
      propagate: true
      handlers: [console, file, error]

    tornado:
      level: WARNING
      propagate: true
      handlers: [console, file, error]

    {{app_name}}:
      level: DEBUG
      propagate: true
      handlers: [console, file, error]

  disable_existing_loggers: true
  incremental: false

Routes:
  - [/, {{app_name}}.DefaultHandler]

"""
        template = template.replace('{{app_name}}', self._arguments.project)
        template = template.replace('{{uuid}}', str(uuid.uuid4()))
        return template

    def _create_package_setup(self):

        template = """from setuptools import setup
import os
from platform import python_version_tuple

requirements = ['tinman']
test_requirements = ['mock', 'nose']
if float('!s.!s' ! python_version_tuple()[0:2]) < 2.7:
    requirements.append('argparse')
    test_requirements.append('unittest2')

# Build the path to install the templates, example config and static files
base_path = '/usr/share/%(project)s'
data_files = dict()
data_paths = ['static', 'templates', 'etc']
for data_path in data_paths:
    for dir_path, dir_names, file_names in os.walk(data_path):
        install_path = '!s/!s' ! (base_path, dir_path)
        if install_path not in data_files:
            data_files[install_path] = list()
        for file_name in file_names:
            data_files[install_path].append('!s/!s' ! (dir_path, file_name))
with open('MANIFEST.in', 'w') as handle:
    for path in data_files:
        for filename in data_files[path]:
            handle.write('include !s\\n' ! filename)


setup(name='%(project)s',
      version='1.0.0',
      packages=['%(project)s'],
      install_requires=requirements,
      test_suite='nose.collector',
      tests_require=test_requirements,
      data_files=[(key, data_files[key]) for key in data_files.keys()],
      zip_safe=True)

"""
        setup_py = template % {'project': self._arguments.project}
        with open('%s/setup.py' % self._arguments.project, 'w') as init:
            init.write(setup_py.replace('!', '%'))

    def _create_subdirectory(self, subdir):
        os.mkdir('%s/%s' % (self._arguments.project, subdir), self.DEFAULT_MODE)

    def _new_argument_parser(self):
        """Return a new argument parser.

        :rtype: argparse.ArgumentParser

        """
        return argparse.ArgumentParser(prog='tinman-init',
                                       conflict_handler='resolve',
                                       description=DESCRIPTION)


    def run(self):
        self._arguments = self._parser.parse_args()
        self._create_directories()
        self._create_package_init()
        self._create_package_setup()
        self._create_config()


def main():
    initializer = Project()
    initializer.run()


if __name__ == '__main__':
    main()



import os
from platform import python_version_tuple
from setuptools import setup
import sys

requirements = ['helper', 'pyyaml', 'tornado>=3.0']
test_requirements = ['mock', 'nose']
(major, minor, rev) = python_version_tuple()
if float('%s.%s' % (major, minor)) < 2.7:
    requirements.append('importlib')
    test_requirements.append('unittest2')

# Build the path to install the templates, example config and static files
if hasattr(sys, 'real_prefix'):
    base_path = 'share/tinman'
else:
    base_path = '/usr/share/tinman'

data_files = {'%s/' % base_path: ['README.md', 'etc/example.yaml'],
              '%s/init.d/' % base_path: ['etc/init.d/tinman'],
              '%s/sysconfig/' % base_path: ['etc/sysconfig/tinman']}

with open('MANIFEST.in', 'w') as handle:
    for path in data_files:
        for filename in data_files[path]:
            handle.write('include %s\n' % filename)

setup(name='tinman',
      version='0.10.0p1',
      description=("Tornado application wrapper and toolset for Tornado "
                   "development"),
      long_description=('Tinman is a take what you need package designed to '
                        'speed development of Tornado applications. It '
                        'includes an application wrapper and a toolbox of '
                        'decorators and utilities.'),
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
      ],
      keywords='tornado',
      author='Gavin M. Roy',
      author_email='gavinmroy@gmail.com',
      url='http://github.com/gmr/tinman',
      license='BSD',
      packages=['tinman',
                'tinman.auth',
                'tinman.decorators',
                'tinman.handlers',
                'tinman.loaders',
                'tinman.session',
                'tinman.utilities'],
      install_requires=requirements,
      extras_require={'Heapy': 'guppy',
                      'LDAP': 'python-ldap',
                      'MsgPack': 'msgpack',
                      'NewRelic': 'newrelic',
                      'PostgreSQL': 'psycopg2',
                      'RabbitMQ': 'pika',
                      'Redis': 'tornado-redis',
                      'Redis Sessions': 'tornado-redis',
                      'Whitelist': 'ipaddr'},
      test_suite='nose.collector',
      tests_require=test_requirements,
      data_files=[(key, data_files[key]) for key in data_files.keys()],
      entry_points=dict(console_scripts=['tinman=tinman.controller:main',
                                         'tinman-init=tinman.utilities.'
                                         'initialize:main',
                                         'tinman-heap-report=tinman.utilities.'
                                         'heapy_report:main']),
      zip_safe=True)



"""
Unlike Tinman's basic authentication decorator, this one is 
applied to the individual methods inside the RequestHandler. 

See helloworld_digest.py in the examples.

"""
from tornado.web import *
from hashlib import md5

class DigestAuthMixin(object):
    def apply_checksum(self, data):
        return md5(data).hexdigest()

    def apply_digest(self, secret, data):
        return self.apply_checksum(secret + ":" + data)

    def A1(self, algorithm, auth_pass):
        """
         If 'algorithm' is "MD5" or unset, A1 is:
         A1 = unq(username-value) ":" unq(realm-value) ":" passwd

         if 'algorithm' is 'MD5-Sess', A1 is:
         A1 = H( unq(username-value) ":" unq(realm-value) ":" passwd )
          ":" unq(nonce-value) ":" unq(cnonce-value)

        """

        username = self.params["username"]
        if algorithm == 'MD5' or not algorithm:
            return "%s:%s:%s" % (username, self.realm, auth_pass)
        elif algorithm == 'MD5-Sess':
            return self.apply_checksum('%s:%s:%s:%s:%s' % \
                                       (username,
                                       self.realm,
                                       auth_pass,
                                       self.params['nonce'],
                                       self.params['cnonce']))


    def A2(self):
        """
        If the "qop" directive's value is "auth" or is unspecified, then A2 is:
            A2 = Method ":" digest-uri-value
        Else,
            A2 = Method ":" digest-uri-value ":" H(entity-body)

        """
        if self.params['qop'] == 'auth' or not self.params['qop']:
            return self.request.method + ":" + self.request.uri
        elif self.params['qop'] == 'auth-int':
            #print "UNSUPPORTED 'qop' METHOD\n"
            return ":".join([self.request.method,
                             self.request.uri,
                             self.apply_checksum(self.request.body)])
        else:
            print "A2 GOT BAD VALUE FOR 'qop': %s\n" % self.params['qop']

    def response(self, auth_pass):
        if 'qop' in self.params:
            auth_comps = [self.params['nonce'],
                               self.params['nc'],
                               self.params['cnonce'],
                               self.params['qop'],
                               self.apply_checksum(self.A2())]
            return self.apply_digest(self.apply_checksum( \
                                    self.A1(self.params.get('algorithm'),
                                            auth_pass)),
                                     ':'.join(auth_comps))
        else:
            return self.apply_digest(self.apply_checksum( \
                                    self.A1(self.params.get('algorithm'),
                                            auth_pass)),
                                    ':'.join([self.params["nonce"],
                                              self.apply_checksum(self.A2())]))

    def _parse_header(self, authheader):
        try:
            n = len("Digest ")
            authheader = authheader[n:].strip()
            items = authheader.split(", ")
            keyvalues = [i.split("=", 1) for i in items]
            keyvalues = ([(k.strip(), v.strip().replace('"', '')) for
                                                            k, v in keyvalues])
            self.params = dict(keyvalues)
        except:
            self.params = []

    def _create_nonce(self):
        return md5("%d:%s" % (time.time(), self.realm)).hexdigest()

    def createAuthHeader(self):
        self.set_status(401)
        nonce = self._create_nonce()
        self.set_header("WWW-Authenticate",
                        "Digest algorithm=MD5 realm=%s qop=auth nonce=%s" %
                        (self.realm, nonce))
        self.finish()

        return False

    def get_authenticated_user(self, get_creds_callback, realm):
        creds = None
        expected_response = None
        actual_response = None
        auth = None
        if not hasattr(self,'realm'):
            self.realm = realm

        try:
            auth = self.request.headers.get('Authorization')
            if not auth or not auth.startswith('Digest '):
                return self.createAuthHeader()
            else:
                self._parse_header(auth)
                required_params = ['username', 'realm', 'nonce', 'uri',
                                   'response', 'qop', 'nc', 'cnonce']
                for k in required_params:
                    if k not in self.params:
                        print "REQUIRED PARAM %s MISSING\n" % k
                        return self.createAuthHeader()
                    elif not self.params[k]:
                        print "REQUIRED PARAM %s IS NONE OR EMPTY\n" % k
                        return self.createAuthHeader()
                    else:
                        continue

            creds = get_creds_callback(self.params['username'])
            if not creds:
                # the username passed to get_creds_callback didn't
                # match any valid users.
                self.createAuthHeader()
            else:
                expected_response = self.response(creds['auth_password'])
                actual_response = self.params['response']
                print "Expected: %s" % expected_response
                print "Actual: %s" % actual_response

            if expected_response and actual_response:
                if expected_response == actual_response:
                    self._current_user = self.params['username']
                    print ("Digest Auth user '%s' successful for realm '%s'. "
                            "URI: '%s', IP: '%s'" % (self.params['username'],
                                                     self.realm,
                                                     self.request.uri,
                                                     self.request.remote_ip))
                    return True
                else:
                    self.createAuthHeader()

        except Exception as out:
            print "FELL THROUGH: %s\n" % out
            print "AUTH HEADERS: %s" % auth
            print "SELF.PARAMS: ",self.params,"\n"
            print "CREDS: ", creds
            print "EXPECTED RESPONSE: %s" % expected_response
            print "ACTUAL RESPONSE: %s" % actual_response
            return self.createAuthHeader()


def digest_auth(realm, auth_func):
    """A decorator used to protect methods with HTTP Digest authentication.

    """
    def digest_auth_decorator(func):
        def func_replacement(self, *args, **kwargs):
            # 'self' here is the RequestHandler object, which is inheriting
            # from DigestAuthMixin to get 'get_authenticated_user'
            if self.get_authenticated_user(auth_func, realm):
                return func(self, *args, **kwargs)
        return func_replacement
    return digest_auth_decorator



"""The heapy handler gives information about the process's memory stack. It is
slow and will block any asynchronous activity and should be used for debugging
purposes only.

For best results, connect directly to the port of the process you would like to
check.

"""
import guppy
import logging
import re
from tornado import web

LOGGER = logging.getLogger(__name__)
MAX_REFERRER_DEPTH = 4
MAX_ROW_COUNT_PER_LEVEL = 5

REPORT_TOTAL = re.compile('^Partition of a set of ([\d]+) objects\.'
                          ' Total size = ([\d]+) bytes\.')
REPORT_HEADER = re.compile('^ Index  Count   %     Size   % Cumulative  % (.*)',
                           re.MULTILINE)
REPORT_ITEMS = re.compile('^\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+'
                          '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)', re.MULTILINE)


def get_report_data(heapy_obj, depth=1):
    LOGGER.debug('Getting report data at depth %i', depth)
    report = {'total_objects': 0, 'total_bytes': 0, 'rows': []}
    totals = REPORT_TOTAL.findall(str(heapy_obj))
    if totals:
        report['total_objects'], report['total_bytes'] = (int(totals[0][0]),
                                                          int(totals[0][1]))
    items = REPORT_ITEMS.findall(str(heapy_obj))
    for index, row in enumerate(items):
        report['rows'].append({'item': row[-1],
                               'count': {'value': int(row[1]),
                                         'percent': int(row[2])},
                               'size': {'value': int(row[3]),
                                        'percent': int(row[4])},
                               'cumulative': {'value': int(row[5]),
                                              'percent': int(row[6])}})
        if depth < MAX_REFERRER_DEPTH:
            try:
                rows = len(heapy_obj.byrcs[index])
            except IndexError:
                LOGGER.warning('Could not process item at index %i', index)
                report['rows'][index]['error'] = 'Could not get referrers'
                continue
            if rows > MAX_ROW_COUNT_PER_LEVEL:
                rows = MAX_ROW_COUNT_PER_LEVEL
            for referrer_index in range(0, rows):
                report['rows'][index]['referrers'] =\
                    get_report_data(heapy_obj.byrcs[index].referrers.byrcs,
                                    depth + 1)

    header = REPORT_HEADER.findall(str(heapy_obj))
    if header:
        report['title'] = header[0]
    return report


class HeapyRequestHandler(web.RequestHandler):
    """Dumps the heap to a text/plain output."""

    def initialize(self):
        self._heapy = guppy.hpy()

    def get(self):
        heap = self._heapy.heap()
        report = get_report_data(heap.byrcs)
        self.write(report)
        self.finish()



#!/usr/bin/env python
"""Will generate a plaintext report from the JSON document created
by the HeapyRequestHandler.

Usage: tinman-heap-report file.json

"""
import json
import os
import sys


def print_row(row, depth):
    prefix = ''.join([' ' for offset in range(0, depth * 4)])
    item = '%s - %s' % (prefix, row['item'])
    parts = [item.ljust(80),
             ('%(value)s' % (row['count'])).rjust(10),
             (' %(percent)s%%' % (row['count'])).rjust(7),
             ('%(value)s' % (row['size'])).rjust(10),
             (' %(percent)s%%' % (row['size'])).rjust(7)]
    print ''.join(parts)


def main():
    if len(sys.argv) == 1 or not (os.path.exists(sys.argv[1]) and
                                  os.path.isfile(sys.argv[1])):
        print 'Usage: tinman-heap-report heap-file.json\n'
        sys.exit(-1)
    with open(sys.argv[1], "r") as handle:
        report = json.load(handle)
    print ''.join(['Item'.ljust(80), 'Count'.rjust(17), 'Size'.rjust(17)])
    print ''.join(['-' for position in xrange(0, 114)])
    for row in report['rows']:
        print
        print_row(row, 0)
        for child in row['referrers']['rows']:
            print
            print_row(child, 1)
            for grandchild in child['referrers']['rows']:
                print_row(grandchild, 2)

if __name__ == '__main__':
    main()



"""The Tinman Controller class, uses clihelper for most of the main
functionality with regard to configuration, logging and daemoniaztion. Spawns a
tornado.HTTPServer and Application per port using multiprocessing.

"""
import helper
import logging
from helper import parser
import multiprocessing
import os
import signal
import sys
import time
from tornado import version as tornado_version

# Tinman Imports
from tinman import __desc__
from tinman import __version__
from tinman import config
from tinman import process

LOGGER = logging.getLogger(__name__)


class Controller(helper.Controller):
    """Tinman controller is the core application coordinator, responsible for
    spawning and managing children.

    """
    APPNAME = 'Tinman'
    DEFAULT_PORTS = [8900]
    MAX_SHUTDOWN_WAIT = 4
    VERSION = __version__

    def enable_debug(self):
        """If the cli arg for foreground is set, set the configuration option
        for debug.

        """
        if self.args.foreground:
            self.config.application[config.DEBUG] = True

    def insert_paths(self):
        """Inserts a base path into the sys.path list if one is specified in
        the configuration.

        """
        if self.args.path:
            sys.path.insert(0, self.args.path)

        if hasattr(self.config.application, config.PATHS):
            if hasattr(self.config.application.paths, config.BASE):
                sys.path.insert(0, self.config.application.paths.base)

    @property
    def living_children(self):
        """Returns a list of all child processes that are still alive.

        :rtype: list

        """
        return [child for child in self.children if child.is_alive()]

    def configuration_reloaded(self):
        """Send a SIGHUP to child processes"""
        LOGGER.info('Notifying children of new configuration updates')
        self.signal_children(signal.SIGHUP)

    def process(self):
        """Check up on child processes and make sure everything is running as
        it should be.

        """
        children = len(self.living_children)
        LOGGER.debug('%i active child%s',
                     children, '' if children == 1 else 'ren')

    @property
    def ports_to_spawn(self):
        """Return the list of ports to spawn

        :rtype: list

        """
        return (self.config.get(config.HTTP_SERVER, dict()).get(config.PORTS)
                or self.DEFAULT_PORTS)

    def set_base_path(self, value):
        """Munge in the base path into the configuration values

        :param str value: The path value

        """
        if config.PATHS not in self.config.application:
            self.config.application[config.PATHS] = dict()

        if config.BASE not in self.config.application[config.PATHS]:
            self.config.application[config.PATHS][config.BASE] = value

    def setup(self):
        """Additional setup steps."""
        LOGGER.info('Tinman v%s starting up with Tornado v%s',
                    __version__, tornado_version)
        # Setup debugging and paths
        self.enable_debug()
        self.set_base_path(os.getcwd())
        self.insert_paths()

        # Setup child processes
        self.children = list()
        self.manager = multiprocessing.Manager()
        self.namespace = self.manager.Namespace()
        self.namespace.args = self.args
        self.namespace.config = dict(self.config.application)
        self.namespace.logging = self.config.logging
        self.namespace.debug = self.debug
        self.namespace.routes = self.config.get(config.ROUTES)
        self.namespace.server = self.config.get(config.HTTP_SERVER)
        self.spawn_processes()

    def shutdown(self):
        """Send SIGABRT to child processes to instruct them to stop"""
        self.signal_children(signal.SIGABRT)

        # Wait a few iterations when trying to stop children before terminating
        waiting = 0
        while self.living_children:
            time.sleep(0.5)
            waiting += 1
            if waiting == self.MAX_SHUTDOWN_WAIT:
                self.signal_children(signal.SIGKILL)
                break

    def signal_children(self, signum):
        """Send a signal to all children

        :param int signum: The signal to send

        """
        LOGGER.info('Sending signal %i to children', signum)
        for child in self.living_children:
            if child.pid != os.getpid():
                os.kill(child.pid, signum)

    def spawn_process(self, port):
        """Create an Application and HTTPServer for the given port.

        :param int port: The port to listen on
        :rtype: multiprocessing.Process

        """
        return process.Process(name="ServerProcess.%i" % port,
                               kwargs={'namespace': self.namespace,
                                       'port': port})

    def spawn_processes(self):
        """Spawn of the appropriate number of application processes"""
        for port in self.ports_to_spawn:
            process = self.spawn_process(port)
            process.start()
            self.children.append(process)


def main():
    """Invoked by the script installed by setuptools."""
    parser.name('tinman')
    parser.description(__desc__)

    p = parser.get()
    p.add_argument('-p', '--path',
                   action='store',
                   dest='path',
                   help='Path to prepend to the Python system path')

    helper.start(Controller)



"""
Tinman data serializers for use with sessions and other data objects.

"""
import datetime
import json
try:
    import msgpack
except ImportError:
    msgpack = None
import pickle


class Serializer(object):
    """Base data serialization object used by session adapters and other
    classes. To use different data serialization formats, extend this class and
    implement the serialize and deserialize methods.

    """
    def deserialize(self, data):
        """Return the deserialized data.

        :param str data: The data to deserialize
        :rtype: dict
        :raises: NotImplementedError

        """
        raise NotImplementedError

    def serialize(self, data):
        """Return self._data as a serialized string.

        :param str data: The data to serialize
        :rtype: str

        """
        raise NotImplementedError

    def _deserialize_datetime(self, data):
        """Take any values coming in as a datetime and deserialize them

        """
        for key in data:
            if isinstance(data[key], dict):
                if data[key].get('type') == 'datetime':
                    data[key] = \
                        datetime.datetime.fromtimestamp(data[key]['value'])
        return data

    def _serialize_datetime(self, data):
        for key in data.keys():
            if isinstance(data[key], datetime.datetime):
                data[key] = {'type': 'datetime',
                             'value': data[key].strftime('%s')}
        return data


class Pickle(Serializer):
    """Serializes the data in Pickle format"""
    def deserialize(self, data):
        """Return the deserialized data.

        :param str data: The data to deserialize
        :rtype: dict

        """
        if not data:
            return dict()
        return self._deserialize_datetime(pickle.loads(data))

    def serialize(self, data):
        """Return self._data as a serialized string.

        :param str data: The data to serialize
        :rtype: str

        """
        return pickle.dumps(self._serialize_datetime(data))


class JSON(Serializer):
    """Serializes the data in JSON format"""
    def deserialize(self, data):
        """Return the deserialized data.

        :param str data: The data to deserialize
        :rtype: dict

        """
        return self._deserialize_datetime(json.loads(data, encoding='utf-8'))

    def serialize(self, data):
        """Return the data as serialized string.

        :param dict data: The data to serialize
        :rtype: str

        """
        return json.dumps(self._serialize_datetime(data), ensure_ascii=False)


class MsgPack(Serializer):
    """Serializes the data in msgpack format"""

    def deserialize(self, data):
        """Return the deserialized data.

        :param str data: The data to deserialize
        :rtype: dict

        """
        return self._deserialize_datetime(msgpack.loads(data))

    def serialize(self, data):
        """Return the data as serialized string.

        :param dict data: The data to serialize
        :rtype: str

        """
        return msgpack.dumps(self._serialize_datetime(data))
